// Databank with a bounded number of objects stored
//
// Copyright 2012, StatusNet Inc.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

var databank = require('../databank'),
    MemoryDatabank = require('./memorydatabank').MemoryDatabank,
    Step = require('step'),
    Databank = databank.Databank,
    DatabankError = databank.DatabankError,
    AlreadyExistsError = databank.AlreadyExistsError,
    NoSuchThingError = databank.NoSuchThingError;

var BoundedDatabank = function(params) {
    this.schema = params.schema || {};
    this.hardLimit = params.hardLimit || 1024;
    this.softLimit = params.softLimit || (this.hardLimit / 2);
    this.lru = [];
};

BoundedDatabank.prototype = new MemoryDatabank();

BoundedDatabank.prototype.trim = function(onCompletion) {
    var cleanup = function() {
	var toSlice = 
    };

    if (this.lru.length >= this.hardLimit) {
    } else if (this.lru.length >= this.softLimit) {
    } else {
	onCompletion(null);
    }
};

BoundedDatabank.prototype.touch = function(type, id, onCompletion) {
};

BoundedDatabank.prototype.create = function(type, id, value, onCompletion) {
    var that = this;
    Step(
	function() {
	    that.trim(this);
	},
	function(err) {
	    if (err) throw err;
	    MemoryDatabank.prototype.create.apply(that, [type, id, value, this]);
	},
	function(err, newValue) {
	    if (!err) {
		that.touch(type, id);
	    }
	    onCompletion(err, newValue);
	}
    );
};

// Read an existing thing
// type = the type of thing; 'user', 'activity'
// id = a unique ID -- nickname or UUID or URI
// onCompletion(err, value) = function to call on completion

BoundedDatabank.prototype.read = function(type, id, onCompletion) {

};

// Update an existing thing
// type = the type of thing; 'user', 'activity'
// id = a unique ID -- nickname or UUID or URI
// value = the new value of the thing
// onCompletion(err, value) = function to call on completion

BoundedDatabank.prototype.update = function(type, id, value, onCompletion) {
    var bank = this;
    Step(
        function() {
            bank.source.update(type, id, value, this);
        },
        function(err, results) {
            // XXX: make sure to clear cache if not found?
            if (err) throw err;
            bank.cache.save(type, id, results, this);
        },
        function(err, results) {
            // XXX: skip if error is from cache?
            if (err) {
                onCompletion(err, null);
            } else {
                onCompletion(null, results);
            }
        }
    );
};

// Delete an existing thing
// type = the type of thing; 'user', 'activity'
// id = a unique ID -- nickname or UUID or URI
// value = the new value of the thing
// onCompletion(err) = function to call on completion

BoundedDatabank.prototype.del = function(type, id, onCompletion) {
    var bank = this;
    Step(
        function() {
            bank.source.del(type, id, this.parallel());
            bank.cache.del(type, id, this.parallel());
        },
        function(err) {
            onCompletion(err);
        }
    );
};

// Search for things
// type = type of thing
// criteria = map of criteria, with exact matches,
//           like {'subject.id' ='tag =example.org,2011 =evan' }
// onResult(value) = called once per result found
// onCompletion(err) = called once at the end of results

BoundedDatabank.prototype.search = function(type, criteria, onResult, onCompletion) {
    // XXX: is there an efficient way to cache this stuff?
    this.source.search(type, criteria, onResult, onCompletion);
};

// Update an existing thing
// type = the type of thing; 'user', 'activity'
// id = a unique ID -- nickname or UUID or URI
// value = the new value of the thing
// onCompletion(err, value) = function to call on completion

BoundedDatabank.prototype.save = function(type, id, value, onCompletion) {
    var bank = this;
    Step(
        function() {
            bank.source.save(type, id, value, this);
        },
        function(err, results) {
            // XXX: make sure to clear cache if not found?
            if (err) throw err;
            bank.cache.save(type, id, results, this);
        },
        function(err, results) {
            // XXX: skip if error is from cache?
            if (err) {
                onCompletion(err, null);
            } else {
                onCompletion(null, results);
            }
        }
    );
};

// Read a bunch of things from the db
// type = the type of thing; 'user', 'activity'
// id = array of IDs
// onCompletion(err, results) = function to call on completion;
//    err = an error or null if none
//    results = map of id (maybe stringified) to value

BoundedDatabank.prototype.readAll = function(type, ids, onCompletion) {
    var bank = this, allMap = {}, notFound = [];

    Step(
        function() {
            bank.cache.readAll(type, ids, this);
        },
        function(err, cacheMap) {
            var i, id, value;
            // XXX: make sure to clear cache if not found?
            if (err) throw err;
            for (i = 0; i < ids.length; i++) {
                id = ids[i];
                // copy the hits
                if (cacheMap.hasOwnProperty(id)) {
                    if (cacheMap[id] === null) {
                        notFound.push(id);
                    } else {
                        allMap[id] = cacheMap[id];
                    }
                }
            }
            if (notFound.length === 0) {
                onCompletion(null, allMap);
            } else {
                bank.source.readAll(type, notFound, this);
            }
        },
        function(err, sourceMap) {
            var i, id;
            if (err) {
                onCompletion(err, null);
            } else {
                for (i = 0; i < ids.length; i++) {
                    id = ids[i];
                    // copy everything
                    if (sourceMap.hasOwnProperty(id)) {
                        allMap[id] = sourceMap[id];
                    }
                }
                onCompletion(null, allMap);
            }
        }
    );
};

BoundedDatabank.prototype.incr = function(type, id, onCompletion) {
    var bank = this;
    Step(
        function() {
            bank.source.incr(type, id, this);
        },
        function(err, value) {
            if (err) throw err;
            bank.cache.save(type, id, value, this);
        },
        function(err, value) {
            if (err) {
                onCompletion(err, null);
            } else {
                onCompletion(null, value);
            }
        }
    );
};

BoundedDatabank.prototype.decr = function(type, id, onCompletion) {
    var bank = this;
    Step(
        function() {
            bank.source.decr(type, id, this);
        },
        function(err, value) {
            if (err) throw err;
            bank.cache.save(type, id, value, this);
        },
        function(err, value) {
            if (err) {
                onCompletion(err, null);
            } else {
                onCompletion(null, value);
            }
        }
    );
};

BoundedDatabank.prototype.append = function(type, id, toAppend, onCompletion) {
    var bank = this;
    Step(
        function() {
            bank.source.append(type, id, toAppend, this);
        },
        function(err, value) {
            if (err) throw err;
            bank.cache.save(type, id, value, this);
        },
        function(err, value) {
            if (err) {
                onCompletion(err, null);
            } else {
                onCompletion(null, value);
            }
        }
    );
};

BoundedDatabank.prototype.prepend = function(type, id, toPrepend, onCompletion) {
    var bank = this;
    Step(
        function() {
            bank.source.prepend(type, id, toPrepend, this);
        },
        function(err, value) {
            if (err) throw err;
            bank.cache.save(type, id, value, this);
        },
        function(err, value) {
            if (err) {
                onCompletion(err, null);
            } else {
                onCompletion(null, value);
            }
        }
    );
};

BoundedDatabank.prototype.item = function(type, id, index, onCompletion) {
    var bank = this, item = null;
    Step(
        function() {
            bank.cache.item(type, id, index, this);
        },
        function(err, value) {
            if (err) {
                if (err instanceof NoSuchThingError) {
                    // Cache miss; read it all and cache
                    bank.source.read(type, id, this);
                } else {
                    // Some other error
                    onCompletion(err, null);
                }
            } else {
                onCompletion(null, value);
            }
        },
        function(err, full) {
            if (err) throw err;
            item = full[index];
            bank.cache.save(type, id, full, this);
        },
        function(err, value) {
            if (err) {
                onCompletion(err, null);
            } else {
                onCompletion(null, item);
            }
        }
    );
};

BoundedDatabank.prototype.slice = function(type, id, start, length, onCompletion) {
    var bank = this, slice = null;
    Step(
        function() {
            bank.cache.slice(type, id, start, length, this);
        },
        function(err, value) {
            if (err) {
                if (err instanceof NoSuchThingError) {
                    // Cache miss; read it all and cache
                    bank.source.read(type, id, this);
                } else {
                    // Some other error
                    onCompletion(err, null);
                }
            } else {
                onCompletion(null, value);
            }
        },
        function(err, full) {
            if (err) throw err;
            slice = full.slice(start, length);
            bank.cache.save(type, id, full, this);
        },
        function(err, value) {
            if (err) {
                onCompletion(err, null);
            } else {
                onCompletion(null, slice);
            }
        }
    );
};

exports.BoundedDatabank = BoundedDatabank;
